'use strict';

Object.defineProperty(exports, "__esModule", {
	value: true
});

var _namespace = require('../namespace');

var _namespace2 = _interopRequireDefault(_namespace);

var _PageStateManager = require('./state/PageStateManager');

var _PageStateManager2 = _interopRequireDefault(_PageStateManager);

var _ObjectContainer = require('../ObjectContainer');

var _ObjectContainer2 = _interopRequireDefault(_ObjectContainer);

var _Controller = require('../controller/Controller');

var _Controller2 = _interopRequireDefault(_Controller);

var _GenericError = require('../error/GenericError');

var _GenericError2 = _interopRequireDefault(_GenericError);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

_namespace2.default.namespace('ima.page');

/**
 * Factory for page.
 */
class PageFactory {

	/**
  * Factory used by page management classes.
  *
  * @param {ObjectContainer} oc
  */
	constructor(oc) {

		/**
   * The current application object container.
   *
   * @type {ObjectContainer}
   */
		this._oc = oc;
	}

	/**
  * Create new instance of {@linkcode Controller}.
  *
  * @param {(string|function(new:Controller))} controller
  * @return {Controller}
  */
	createController(controller) {
		let controllerInstance = this._oc.create(controller);

		return controllerInstance;
	}

	/**
  * Retrieves the specified react component class.
  *
  * @param {(string|function(new: React.Component))} view The namespace
  *        referring to a react component class, or a react component class
  *        constructor.
  * @return {function(new: React.Component)} The react component class
  *         constructor.
  */
	createView(view) {
		if (typeof view === 'function') {
			return view;
		}
		let classConstructor = this._oc.getConstructorOf(view);

		if (classConstructor) {
			return classConstructor;
		} else {
			throw new _GenericError2.default(`ima.page.Factory:createView hasn't name of view "${view}".`);
		}
	}

	/**
  * Returns decorated controller for ease setting seo params in controller.
  *
  * @param {Controller} controller
  * @return {Controller}
  */
	decorateController(controller) {
		let metaManager = this._oc.get('$MetaManager');
		let router = this._oc.get('$Router');
		let dictionary = this._oc.get('$Dictionary');
		let settings = this._oc.get('$Settings');

		let decoratedController = this._oc.create('$ControllerDecorator', [controller, metaManager, router, dictionary, settings]);

		return decoratedController;
	}

	/**
  * Returns decorated page state manager for extension.
  *
  * @param {PageStateManager} pageStateManager
  * @param {string[]} allowedStateKeys
  * @return {PageStateManager}
  */
	decoratePageStateManager(pageStateManager, allowedStateKeys) {
		let decoratedPageStateManager = this._oc.create('$PageStateManagerDecorator', [pageStateManager, allowedStateKeys]);

		return decoratedPageStateManager;
	}
}

exports.default = PageFactory;
_namespace2.default.ima.page.PageFactory = PageFactory;

typeof $IMA !== 'undefined' && $IMA !== null && $IMA.Loader && $IMA.Loader.register('ima/page/PageFactory', [], function (_export, _context) {
	'use strict';
	return {
		setters: [],
		execute: function () {
			_export('default', exports.default);
		}
	};
});
